package cz.muni.fi.pa165.movie;

import cz.muni.fi.pa165.CoreApplication;
import io.swagger.v3.oas.annotations.Operation;
import io.swagger.v3.oas.annotations.media.Content;
import io.swagger.v3.oas.annotations.responses.ApiResponse;
import io.swagger.v3.oas.annotations.security.SecurityRequirement;
import jakarta.servlet.http.HttpServletResponse;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.stereotype.Controller;
import org.springframework.web.bind.annotation.*;

import javax.validation.Valid;
import java.util.List;

@Controller
@RestController
@RequestMapping(path = "/movies", produces = MediaType.APPLICATION_JSON_VALUE)
public class MovieController {

    private final MovieService movieService;

    @Autowired
    public MovieController(MovieService movieService) {
        this.movieService = movieService;
    }


    @Operation(summary = "Get all movies",
            security = @SecurityRequirement(name = CoreApplication.SECURITY_SCHEME_NAME, scopes = {"test_read"}),
            responses = {
                    @ApiResponse(responseCode = "200", description = "OK - movies were successfully returned"),
                    @ApiResponse(responseCode = "401", description = "Unauthorized - access token not provided or valid", content = @Content()),
                    @ApiResponse(responseCode = "403", description = "Forbidden - access token does not have scope test_read", content = @Content()),
            }
    )
    @GetMapping
    public List<MovieDto> getMovies() {
        return movieService.getMovies();
    }

    @Operation(summary = "Get movie by id",
            security = @SecurityRequirement(name = CoreApplication.SECURITY_SCHEME_NAME, scopes = {"test_read"}),
            responses = {
                    @ApiResponse(responseCode = "200", description = "OK - movie was successfully returned"),
                    @ApiResponse(responseCode = "401", description = "Unauthorized - access token not provided or valid", content = @Content()),
                    @ApiResponse(responseCode = "403", description = "Forbidden - access token does not have scope test_read", content = @Content()),
            }
    )
    @GetMapping("{id}")
    public MovieDto getMovie(@PathVariable long id) {
        return movieService.getMovie(id);
    }

    @Operation(summary = "Create new movie",
            security = @SecurityRequirement(name = CoreApplication.SECURITY_SCHEME_NAME, scopes = {"test_write"}),
            responses = {
                    @ApiResponse(responseCode = "200", description = "OK - movie was successfully created"),
                    @ApiResponse(responseCode = "401", description = "Unauthorized - access token not provided or valid", content = @Content()),
                    @ApiResponse(responseCode = "403", description = "Forbidden - access token does not have scope test_write", content = @Content()),
            }
    )
    @PostMapping
    public void createMovie(@Valid @RequestBody MovieDto movieDto) {
        movieService.createMovie(movieDto);
    }


    @Operation(summary = "Update movie",
            security = @SecurityRequirement(name = CoreApplication.SECURITY_SCHEME_NAME, scopes = {"test_write"}),
            responses = {
                    @ApiResponse(responseCode = "200", description = "OK - movie was successfully updated"),
                    @ApiResponse(responseCode = "401", description = "Unauthorized - access token not provided or valid", content = @Content()),
                    @ApiResponse(responseCode = "403", description = "Forbidden - access token does not have scope test_write", content = @Content()),
            }
    )
    @PostMapping("{id}")
    public void updateMovie(@PathVariable long id, @Valid @RequestBody MovieDto movieDto) {
        movieService.updateMovie(id, movieDto);
    }

    @Operation(summary = "Delete movie with id",
            security = @SecurityRequirement(name = CoreApplication.SECURITY_SCHEME_NAME, scopes = {"test_1"}),
            responses = {
                    @ApiResponse(responseCode = "200", description = "OK - movie was successfully deleted"),
                    @ApiResponse(responseCode = "401", description = "Unauthorized - access token not provided or valid", content = @Content()),
                    @ApiResponse(responseCode = "403", description = "Forbidden - access token does not have scope test_1", content = @Content()),
            }
    )
    @DeleteMapping("{id}")
    public void deleteMovie(@PathVariable long id) {
        movieService.deleteMovie(id);
    }

    @Operation(summary = "Delete all movies",
            security = @SecurityRequirement(name = CoreApplication.SECURITY_SCHEME_NAME, scopes = {"test_1"}),
            responses = {
                    @ApiResponse(responseCode = "200", description = "OK - movies were successfully deleted"),
                    @ApiResponse(responseCode = "401", description = "Unauthorized - access token not provided or valid", content = @Content()),
                    @ApiResponse(responseCode = "403", description = "Forbidden - access token does not have scope test_1", content = @Content()),
            }
    )
    @DeleteMapping
    public void deleteAllMovies() {
        movieService.deleteAllMovies();
    }


    @Operation(summary = "Get movie by title",
            security = @SecurityRequirement(name = CoreApplication.SECURITY_SCHEME_NAME, scopes = {"test_read"}),
            responses = {
                    @ApiResponse(responseCode = "200", description = "OK - movie was successfully returned"),
                    @ApiResponse(responseCode = "401", description = "Unauthorized - access token not provided or valid", content = @Content()),
                    @ApiResponse(responseCode = "403", description = "Forbidden - access token does not have scope test_read", content = @Content()),
            }
    )
    @GetMapping("/title/{title}")
    public MovieDto findByTitle(@PathVariable String title) {
        return movieService.findByTitle(title);
    }
}
