package cz.muni.fi.pa156;

import com.fasterxml.jackson.databind.ObjectMapper;
import cz.muni.fi.pa165.AppConfig;
import cz.muni.fi.pa165.RecommenderApplication;
import cz.muni.fi.pa165.RecommenderService;
import cz.muni.fi.pa165.movie.GenreEnum;
import cz.muni.fi.pa165.movie.MovieDto;
import cz.muni.fi.pa165.rating.RatingCategoryEnum;
import cz.muni.fi.pa165.rating.RatingDto;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.autoconfigure.web.servlet.AutoConfigureMockMvc;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.context.annotation.Import;
import org.springframework.http.HttpMethod;
import org.springframework.http.MediaType;
import org.springframework.test.context.ActiveProfiles;
import org.springframework.test.context.junit.jupiter.SpringExtension;
import org.springframework.test.web.client.MockRestServiceServer;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.web.client.RestTemplate;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import static org.assertj.core.api.Assertions.assertThat;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.springframework.test.web.client.match.MockRestRequestMatchers.method;
import static org.springframework.test.web.client.match.MockRestRequestMatchers.requestTo;
import static org.springframework.test.web.client.response.MockRestResponseCreators.withSuccess;


@ExtendWith(SpringExtension.class)
@SpringBootTest(classes = RecommenderApplication.class)
@AutoConfigureMockMvc
@Import(AppConfig.class)
@ActiveProfiles("test")
public class RecommenderServiceTest {

    private MockMvc mockMvc;
    private RestTemplate restTemplate;
    private RecommenderService recommenderService;

    @Autowired
    public RecommenderServiceTest(RestTemplate restTemplate, MockMvc mockMvc, RecommenderService recommenderService) {
        this.restTemplate = restTemplate;
        this.mockMvc = mockMvc;
        this.recommenderService = recommenderService;
    }

    @Test
    public void testGetSimilarMoviesWithNoMovies() throws Exception {
        int targetMovieId = 20;
        String movieControllerUrl = "http://localhost:8090/movies";

        String moviesDtoJson = "[]";

        MockRestServiceServer mockServer = MockRestServiceServer.createServer(restTemplate);
        mockServer.expect(requestTo(movieControllerUrl))
                .andExpect(method(HttpMethod.GET))
                .andRespond(withSuccess(moviesDtoJson, MediaType.APPLICATION_JSON));

        List<MovieDto> similarMovies = recommenderService.getSimilarMovies(targetMovieId, null);

        assertTrue(similarMovies.isEmpty());

        mockServer.verify();
    }

    @Test
    public void testGetSimilarMoviesWithNoRatings() throws Exception {
        int targetMovieId = 20;
        String movieControllerUrl = "http://localhost:8090/movies";
        String ratingControllerUrl = "http://localhost:8090/ratings";

        String moviesDtoJson = "[{\"id\":1,\"title\":\"Title\",\"description\":\"Description\",\"director\":\"Director\",\"actors\":null,\"genres\":null,\"images\":null}]";
        String ratingsDtoJson = "[]";

        MockRestServiceServer mockServer = MockRestServiceServer.createServer(restTemplate);
        mockServer.expect(requestTo(movieControllerUrl))
                .andExpect(method(HttpMethod.GET))
                .andRespond(withSuccess(moviesDtoJson, MediaType.APPLICATION_JSON));

        mockServer.expect(requestTo(ratingControllerUrl))
                .andExpect(method(HttpMethod.GET))
                .andRespond(withSuccess(ratingsDtoJson, MediaType.APPLICATION_JSON));

        List<MovieDto> similarMovies = recommenderService.getSimilarMovies(targetMovieId, null);

        assertTrue(similarMovies.isEmpty());

        mockServer.verify();
    }


    @Test
    public void testGetSimilarMoviesWithInvalidMovieId() throws Exception {
        long targetMovieId = 12345;
        String movieControllerUrl = "http://localhost:8090/movies";
        String ratingControllerUrl = "http://localhost:8090/ratings";

        ObjectMapper objectMapper = new ObjectMapper();

        String moviesDtoJson = objectMapper.writeValueAsString(createMoviesForTesting());
        String ratingsDtoJson = objectMapper.writeValueAsString(createRatingsForTesting());

        MockRestServiceServer mockServer = MockRestServiceServer.createServer(restTemplate);
        mockServer.expect(requestTo(movieControllerUrl))
                .andExpect(method(HttpMethod.GET))
                .andRespond(withSuccess(moviesDtoJson, MediaType.APPLICATION_JSON));

        mockServer.expect(requestTo(ratingControllerUrl))
                .andExpect(method(HttpMethod.GET))
                .andRespond(withSuccess(ratingsDtoJson, MediaType.APPLICATION_JSON));

        List<MovieDto> similarMovies = recommenderService.getSimilarMovies(targetMovieId, null);

        assertTrue(similarMovies.isEmpty());

        mockServer.verify();
    }

    @Test
    public void testGetSimilarMoviesWithExistingTargetMovie() throws Exception {
        int targetMovieId = 10;
        String movieControllerUrl = "http://localhost:8090/movies";
        String ratingControllerUrl = "http://localhost:8090/ratings";

        ObjectMapper objectMapper = new ObjectMapper();

        String moviesDtoJson = objectMapper.writeValueAsString(createMoviesForTesting());
        String ratingsDtoJson = objectMapper.writeValueAsString(createRatingsForTesting());

        MockRestServiceServer mockServer = MockRestServiceServer.createServer(restTemplate);
        mockServer.expect(requestTo(movieControllerUrl))
                .andExpect(method(HttpMethod.GET))
                .andRespond(withSuccess(moviesDtoJson, MediaType.APPLICATION_JSON));

        mockServer.expect(requestTo(ratingControllerUrl))
                .andExpect(method(HttpMethod.GET))
                .andRespond(withSuccess(ratingsDtoJson, MediaType.APPLICATION_JSON));

        List<MovieDto> similarMovies = recommenderService.getSimilarMovies(targetMovieId, null);

        assertThat(similarMovies).isEqualTo(getExpectedMoviesAsResult());

        mockServer.verify();
    }

    @Test
    void testCalculateRatingSimilarityValue() {
        RatingDto rating1 = new RatingDto();
        rating1.setRating(5);
        RatingDto rating2 = new RatingDto();
        rating2.setRating(3);
        rating2.setMovieId(2);

        double similarityScore = recommenderService.getRatingsSimilarityForTesting(List.of(rating1), List.of(rating2), rating2.getMovieId());

        assertTrue(similarityScore >= 0 && similarityScore <= 1);
    }

    @Test
    void testCalculateGenreSimilarityValue() {
        MovieDto movie1 = new MovieDto();
        movie1.setGenres(List.of(GenreEnum.ACTION, GenreEnum.THRILLER));
        MovieDto movie2 = new MovieDto();
        movie2.setGenres(List.of(GenreEnum.ACTION, GenreEnum.SPORT));

        double similarityScore = recommenderService.getGenresSimilarityForTesting(movie1, movie2);

        assertTrue(similarityScore >= 0 && similarityScore <= 1);
    }

    private List<MovieDto> createMoviesForTesting() {
        List<MovieDto> movies = new ArrayList<>();

        MovieDto movie1 = new MovieDto();
        movie1.setId(1L);
        movie1.setTitle("The Godfather");
        movie1.setDescription("A classic crime drama");
        movie1.setDirector("Francis Ford Coppola");
        movie1.setGenres(Arrays.asList(GenreEnum.SCIFI, GenreEnum.DRAMA));
        movies.add(movie1);

        MovieDto movie2 = new MovieDto();
        movie2.setId(2L);
        movie2.setTitle("The Godfather: Part II");
        movie2.setDescription("A sequel to the classic crime drama");
        movie2.setDirector("Francis Ford Coppola");
        movie2.setGenres(Arrays.asList(GenreEnum.SCIFI, GenreEnum.DRAMA));
        movies.add(movie2);

        MovieDto movie3 = new MovieDto();
        movie3.setId(3L);
        movie3.setTitle("The Dark Knight");
        movie3.setDescription("A great action movie");
        movie3.setDirector("Christopher Nolan");
        movie3.setGenres(Arrays.asList(GenreEnum.ACTION, GenreEnum.THRILLER));
        movies.add(movie3);

        MovieDto movie4 = new MovieDto();
        movie4.setId(4L);
        movie4.setTitle("Inception");
        movie4.setDescription("A mind-bending thriller");
        movie4.setDirector("Christopher Nolan");
        movie4.setGenres(Arrays.asList(GenreEnum.ACTION, GenreEnum.SCIFI, GenreEnum.THRILLER));
        movies.add(movie4);

        MovieDto movie5 = new MovieDto();
        movie5.setId(5L);
        movie5.setTitle("The Shawshank Redemption");
        movie5.setDescription("A prison drama");
        movie5.setDirector("Frank Darabont");
        movie5.setGenres(Arrays.asList(GenreEnum.DRAMA));
        movies.add(movie5);

        MovieDto movie6 = new MovieDto();
        movie6.setId(6L);
        movie6.setTitle("Pulp Fiction");
        movie6.setDescription("A classic crime movie");
        movie6.setDirector("Quentin Tarantino");
        movie6.setGenres(Arrays.asList(GenreEnum.SCIFI, GenreEnum.DRAMA));
        movies.add(movie6);

        MovieDto movie7 = new MovieDto();
        movie7.setId(7L);
        movie7.setTitle("The Matrix");
        movie7.setDescription("A sci-fi action movie");
        movie7.setDirector("The Wachowskis");
        movie7.setGenres(Arrays.asList(GenreEnum.ACTION, GenreEnum.SCIFI));
        movies.add(movie7);

        MovieDto movie8 = new MovieDto();
        movie8.setId(8L);
        movie8.setTitle("The Silence of the Lambs");
        movie8.setDescription("A psychological thriller");
        movie8.setDirector("Jonathan Demme");
        movie8.setGenres(Arrays.asList(GenreEnum.SCIFI, GenreEnum.DRAMA, GenreEnum.THRILLER));
        movies.add(movie8);

        MovieDto movie9 = new MovieDto();
        movie9.setId(9L);
        movie9.setTitle("The Lord of the Rings: The Fellowship of the Ring");
        movie9.setDescription("An epic fantasy adventure");
        movie9.setDirector("Peter Jackson");
        movie9.setGenres(Arrays.asList(GenreEnum.FANTASY, GenreEnum.ACTION));
        movies.add(movie9);

        MovieDto movie10 = new MovieDto();
        movie10.setId(10L);
        movie10.setTitle("The Avengers");
        movie10.setDescription("A superhero team-up movie");
        movie10.setDirector("Joss Whedon");
        movie10.setGenres(Arrays.asList(GenreEnum.ACTION, GenreEnum.SCIFI));
        movies.add(movie10);
        return movies;
    }

    private List<MovieDto> getExpectedMoviesAsResult() {
        List<MovieDto> movies = new ArrayList<>();

        MovieDto movie7 = new MovieDto();
        movie7.setId(7L);
        movie7.setTitle("The Matrix");
        movie7.setDescription("A sci-fi action movie");
        movie7.setDirector("The Wachowskis");
        movie7.setGenres(Arrays.asList(GenreEnum.ACTION, GenreEnum.SCIFI));
        movies.add(movie7);

        MovieDto movie4 = new MovieDto();
        movie4.setId(4L);
        movie4.setTitle("Inception");
        movie4.setDescription("A mind-bending thriller");
        movie4.setDirector("Christopher Nolan");
        movie4.setGenres(Arrays.asList(GenreEnum.ACTION, GenreEnum.SCIFI, GenreEnum.THRILLER));
        movies.add(movie4);

        MovieDto movie1 = new MovieDto();
        movie1.setId(1L);
        movie1.setTitle("The Godfather");
        movie1.setDescription("A classic crime drama");
        movie1.setDirector("Francis Ford Coppola");
        movie1.setGenres(Arrays.asList(GenreEnum.SCIFI, GenreEnum.DRAMA));
        movies.add(movie1);

        MovieDto movie2 = new MovieDto();
        movie2.setId(2L);
        movie2.setTitle("The Godfather: Part II");
        movie2.setDescription("A sequel to the classic crime drama");
        movie2.setDirector("Francis Ford Coppola");
        movie2.setGenres(Arrays.asList(GenreEnum.SCIFI, GenreEnum.DRAMA));
        movies.add(movie2);

        MovieDto movie6 = new MovieDto();
        movie6.setId(6L);
        movie6.setTitle("Pulp Fiction");
        movie6.setDescription("A classic crime movie");
        movie6.setDirector("Quentin Tarantino");
        movie6.setGenres(Arrays.asList(GenreEnum.SCIFI, GenreEnum.DRAMA));
        movies.add(movie6);

        return movies;
    }

    private List<RatingDto> createRatingsForTesting() {
        List<RatingDto> ratings = new ArrayList<>();

        RatingDto rating1 = new RatingDto();
        rating1.setId(1L);
        rating1.setCategory(RatingCategoryEnum.PLOT);
        rating1.setRating(4);
        rating1.setUserId(1);
        rating1.setMovieId(10);
        ratings.add(rating1);

        RatingDto rating2 = new RatingDto();
        rating2.setId(2L);
        rating2.setCategory(RatingCategoryEnum.CAST);
        rating2.setRating(5);
        rating2.setUserId(1);
        rating2.setMovieId(7);
        ratings.add(rating2);

        RatingDto rating3 = new RatingDto();
        rating3.setId(3L);
        rating3.setCategory(RatingCategoryEnum.PLOT);
        rating3.setRating(3);
        rating3.setUserId(1);
        rating3.setMovieId(3);
        ratings.add(rating3);

        RatingDto rating4 = new RatingDto();
        rating4.setId(4L);
        rating4.setCategory(RatingCategoryEnum.CAST);
        rating4.setRating(4);
        rating4.setUserId(1);
        rating4.setMovieId(9);
        ratings.add(rating4);

        RatingDto rating5 = new RatingDto();
        rating5.setId(5L);
        rating5.setCategory(RatingCategoryEnum.CAST);
        rating5.setRating(5);
        rating5.setUserId(2);
        rating5.setMovieId(4);
        ratings.add(rating5);

        RatingDto rating6 = new RatingDto();
        rating6.setId(6L);
        rating6.setCategory(RatingCategoryEnum.PLOT);
        rating6.setRating(3);
        rating6.setUserId(2);
        rating6.setMovieId(2);
        ratings.add(rating6);

        RatingDto rating7 = new RatingDto();
        rating7.setId(7L);
        rating7.setCategory(RatingCategoryEnum.CAST);
        rating7.setRating(4);
        rating7.setUserId(2);
        rating7.setMovieId(2);
        ratings.add(rating7);

        RatingDto rating8 = new RatingDto();
        rating8.setId(8L);
        rating8.setCategory(RatingCategoryEnum.OVERALL);
        rating8.setRating(5);
        rating8.setUserId(3);
        rating8.setMovieId(3);
        ratings.add(rating8);

        RatingDto rating9 = new RatingDto();
        rating9.setId(9L);
        rating9.setCategory(RatingCategoryEnum.FUN);
        rating9.setRating(2);
        rating9.setUserId(3);
        rating9.setMovieId(8);
        ratings.add(rating9);

        RatingDto rating10 = new RatingDto();
        rating10.setId(10L);
        rating10.setCategory(RatingCategoryEnum.SPECIAL_EFFECTS);
        rating10.setRating(4);
        rating10.setUserId(4);
        rating10.setMovieId(6);
        ratings.add(rating10);

        RatingDto rating11 = new RatingDto();
        rating11.setId(11L);
        rating11.setCategory(RatingCategoryEnum.OVERALL);
        rating11.setRating(5);
        rating11.setUserId(4);
        rating11.setMovieId(5);
        ratings.add(rating11);

        RatingDto rating12 = new RatingDto();
        rating12.setId(12L);
        rating12.setCategory(RatingCategoryEnum.PLOT);
        rating12.setRating(4);
        rating12.setUserId(4);
        rating12.setMovieId(1);
        ratings.add(rating12);

        RatingDto rating13 = new RatingDto();
        rating13.setId(13L);
        rating13.setCategory(RatingCategoryEnum.CAST);
        rating13.setRating(3);
        rating13.setUserId(5);
        rating13.setMovieId(1);
        ratings.add(rating13);


        RatingDto rating14 = new RatingDto();
        rating14.setId(14L);
        rating14.setCategory(RatingCategoryEnum.OVERALL);
        rating14.setRating(5);
        rating14.setUserId(5);
        rating14.setMovieId(9);
        ratings.add(rating14);

        RatingDto rating15 = new RatingDto();
        rating15.setId(15L);
        rating15.setCategory(RatingCategoryEnum.PLOT);
        rating15.setRating(3);
        rating15.setUserId(5);
        rating15.setMovieId(10);
        ratings.add(rating15);

        RatingDto rating16 = new RatingDto();
        rating16.setId(16L);
        rating16.setCategory(RatingCategoryEnum.FUN);
        rating16.setRating(4);
        rating16.setUserId(6);
        rating16.setMovieId(4);
        ratings.add(rating16);

        RatingDto rating17 = new RatingDto();
        rating17.setId(17L);
        rating17.setCategory(RatingCategoryEnum.CAST);
        rating17.setRating(5);
        rating17.setUserId(6);
        rating17.setMovieId(9);
        ratings.add(rating17);

        RatingDto rating18 = new RatingDto();
        rating18.setId(18L);
        rating18.setCategory(RatingCategoryEnum.SPECIAL_EFFECTS);
        rating18.setRating(2);
        rating18.setUserId(7);
        rating18.setMovieId(1);
        ratings.add(rating18);

        RatingDto rating19 = new RatingDto();
        rating19.setId(19L);
        rating19.setCategory(RatingCategoryEnum.OVERALL);
        rating19.setRating(5);
        rating19.setUserId(7);
        rating19.setMovieId(10);
        ratings.add(rating19);

        RatingDto rating20 = new RatingDto();
        rating20.setId(20L);
        rating20.setCategory(RatingCategoryEnum.PLOT);
        rating20.setRating(4);
        rating20.setUserId(7);
        rating20.setMovieId(9);
        ratings.add(rating20);

        return ratings;
    }
}
