package cz.muni.fi.pa165;

import cz.muni.fi.pa165.movie.GenreEnum;
import cz.muni.fi.pa165.movie.MovieDto;
import cz.muni.fi.pa165.rating.RatingDto;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.core.ParameterizedTypeReference;
import org.springframework.http.HttpMethod;
import org.springframework.http.ResponseEntity;
import org.springframework.stereotype.Service;
import org.springframework.web.client.RestTemplate;

import java.io.*;
import java.util.List;

@Service
public class ReportService {

    @Value("${CORE_APP_URL:http://localhost:8080}")
    private String coreAppUrl;

    private final RestTemplate restTemplate;

    @Autowired
    public ReportService(RestTemplate restTemplate) {
        this.restTemplate = restTemplate;
    }

    public byte[] getMoviesWithAverageRating(String fileName) throws IOException {
        String moviesUrl = coreAppUrl + "/movies";
        String ratingsUrl = coreAppUrl + "/ratings";

        List<MovieDto> allMovies = getMovies(moviesUrl);
        if (allMovies == null || allMovies.isEmpty()) {
            return null;
        }

        List<RatingDto> allRatings = getRatings(ratingsUrl);
        if (allRatings == null || allRatings.isEmpty()) {
            return null;
        }

        return createFileWithAverageRatings(allMovies, allRatings, fileName);
    }

    public byte[] getMoviesWithAverageRatingByGenre(GenreEnum genre, String fileName) throws IOException {
        String moviesUrl = coreAppUrl + "/movies";
        String ratingsUrl = coreAppUrl + "/ratings";

        List<MovieDto> allMovies = getMovies(moviesUrl);
        if (allMovies == null || allMovies.isEmpty()) {
            return null;
        }

        List<RatingDto> allRatings = getRatings(ratingsUrl);
        if (allRatings == null || allRatings.isEmpty()) {
            return null;
        }

        List<MovieDto> moviesByGenre = allMovies.stream().filter(movie -> movie.getGenres().contains(genre)).toList();
        return createFileWithAverageRatings(moviesByGenre, allRatings, fileName);
    }

    private byte[] createFileWithAverageRatings(List<MovieDto> allMovies, List<RatingDto> allRatings, String fileName) throws IOException {
        List<RatedMovie> ratedMovies = allMovies.stream()
                .map(movie -> {
                    List<RatingDto> movieRatings = allRatings.stream().filter(rating -> rating.getMovieId() == movie.getId()).toList();
                    double averageRating = calculateAverageRating(movieRatings);
                    return new RatedMovie(movie, averageRating);
                }).sorted().toList();

        ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
        OutputStreamWriter writer = new OutputStreamWriter(outputStream);

        for (RatedMovie ratedMovie : ratedMovies) {
            writer.write(ratedMovie.movie.getTitle() + " (" + ratedMovie.movie.getDirector() + ") \n" +
                    "   average rating: " + String.format("%.2f", ratedMovie.averageRating) + "/10\n" +
                    "   genres: " + ratedMovie.movie.getGenres() + "\n" +
                    "   actors: " + ratedMovie.movie.getActors() + "\n");
        }

        writer.flush();
        return outputStream.toByteArray();
    }

    private List<RatingDto> getRatings(String ratingsUrl) {
        ResponseEntity<List<RatingDto>> ratingsResponse = restTemplate.exchange(
                ratingsUrl,
                HttpMethod.GET,
                null,
                new ParameterizedTypeReference<>() {
                });
        return ratingsResponse.getBody();
    }

    private List<MovieDto> getMovies(String moviesUrl) {
        ResponseEntity<List<MovieDto>> moviesResponse = restTemplate.exchange(
                moviesUrl,
                HttpMethod.GET,
                null,
                new ParameterizedTypeReference<>() {
                });
        return moviesResponse.getBody();
    }

    private record RatedMovie(MovieDto movie, double averageRating) implements Comparable<RatedMovie> {
        @Override
        public int compareTo(RatedMovie other) {
            return Double.compare(other.averageRating, this.averageRating);
        }
    }

    private double calculateAverageRating(List<RatingDto> ratings) {
        double sumRatings = ratings.stream()
                .mapToDouble(RatingDto::getRating)
                .sum();

        if (ratings.isEmpty()) {
            return 0;
        }
        return sumRatings / ratings.size();
    }
}
